//Main implementation file for XBob4 Encoder Interface

/*	BOB-4 SEI Smart Encoder Interface Firmware Version 1.4
	Processor	AVR ATMega88
	Clock		9.8304 Mhz
	Compiler	AVR-GCC
	Project		AVR Studio
	Author 		Eric L. Jennings
	Copyright	Decade Engineering 2007
	Compiled	Jan. 16 2008

	Version History

	1.0 7/23/2007
		Initial release
	1.1 11/14/2007
		various bug fixes
		reduced startup delay
		added code to clear BOB-4 bootscript
	1.2 11/28/2007
		fixed zero display bug
		changed startup message to conform with status report
	1.3 11/29/2007
		decreased fast count time for increment/decrement switches
	1.4 01/16/08
		switched to fast ASM count interrupt routines
	1.5 8/22/2008
		bug fixes 
		fixed start up count restore
	1.6 6/4/09
		fixed refresh rate slowdown bug
	 
 
Program flow
operates in SPI master mode
count is updated automatically by encoder ISRs located in fastencoder.s
when count changes new count is sent to BOB-4 via SPI
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h> 
#include <avr/eeprom.h>
#include "XBOB4EncInt.h"
#include <util/delay.h>

#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "uart.h"
#include "spi.h"
#include "switch.h"
#include "parser.h"

int32_t count;
int32_t lastcount;
uint8_t metric;
uint8_t screen;
uint8_t poscmd[9];
uint8_t	clrcmd[5];
uint8_t mask;
uint16_t scaler;
uint8_t labelf;
uint8_t labelm;
uint8_t width;
uint8_t mode0x;
uint8_t mode0y;
uint8_t mode1x;
uint8_t mode1y;
uint8_t mode2x;
uint8_t mode2y;
uint8_t mode3x;
uint8_t mode3y;
uint8_t mode4x;
uint8_t mode4y;
double scalefactor;

volatile register uint8_t byte0 asm("r2");
volatile register uint8_t byte1 asm("r3");
volatile register uint8_t byte2 asm("r4");
volatile register uint8_t byte3 asm("r5");
volatile register uint8_t lastvalue asm("r6");


void get_count(void);
void set_count(void);
/**********************************************/

ISR(PCINT2_vect)
{
	//this interrupt is triggered by change on PD6 indicating powerdown
	//so we save the count to EEPROM
	//first shut off indicator LEDs
	PORTB |= (1<<PB0)|(1<<PB1);
	eeprom_write_block (&count, (uint8_t *)EECOUNT, 4);
	PutChar('*');
}	
/**********************************************/

void MCUInit(void)
{
	//set up encoder interrupts on any change
	EICRA = (1<<ISC00)|(1<<ISC10);
	//enable external interrupts for encoder
 	EIMSK = (1<<INT0)|(1<<INT1);
	//enable pin change interrupt for PD6 power fail indicator
	PCICR = (1<<PCIE2);
	PCMSK2 = (1<<PCINT22);
	//global interrupt enable
	//set up PB0/PB1 as outputs for encoder indicators
	DDRB |= (1<<PB0)|(1<<PB1);
	PORTB |= (1<<PB0)|(1<<PB1);
	//global interupt enable
	sei();
}
/**********************************************/

void Refresh(void)
{
	//refresh the screen
	//first send position command
	SpiSendString(poscmd);
	//format and send count to BOB-4
	CountToAscii((double)count);
}

/**********************************************/
void CountToAscii(double value)
{
	//apply scaling, metric conversion (if applicable),format, and send count to BOB-4 for display
	uint8_t countstr[16];
	uint8_t label = labelf;
	memset (countstr,0,16);

	//apply metric conversion
	if(metric)
	{
		value = CountToMetric(value);
		label = labelm;
	}

	//apply scale factor
	value *= scalefactor;

	//convert to a string 
	dtostrf(value,width,mask,countstr);

	//convert leading spaces to leading zeros
	uint8_t index = 1;
	while (countstr[index] != 0)
	{
		if (countstr[index] == '-')
		{	
			countstr[index] = '0';
		}
		else if (countstr[index] == ' ')
		{
			countstr[index] = '0';
		}
		index++;
	}

	//limit string to width
	uint8_t i;
	while (index > width)
	{
		for (i = 1;i < 16;i++)
		{
			countstr[i] = countstr[i+1];
		}
		index--;	
	}

	//add polarity symbol			
	if (value == 0)
		countstr[0] = ' ';
	if (value < 0)
		countstr[0] = '-';
	if (value > 0)
		countstr[0] = '+';
	
	//send to BOB-4
	SpiSendString(countstr);
	
	//add units label
	SpiPutChar(label);				
}
/**********************************************/

//multiply by .3048 to convert feet to metric
double CountToMetric(double val)
{	
	int32_t ret;
	ret = val * .3048;
	return(ret);
}
/**********************************************/


int main(void)
{
	//wait 100 milliseconds for BOB4 initialization and power supply to stabilize 
	uint8_t index;
	for (index = 0;index < 10;index++)
	{
		_delay_ms(10);
	}
	
	//initialize systems		
	metric = FALSE;
	UartInit();
	MCUInit();
	SpiInit();
	SwitchInit();

	//"Greetings Professor Falken, would you like to play a game?"
	SendString(VERSION_STRING);
	PutChar(13);

	clrcmd[0] = 0x1b;
	clrcmd[1] = '[';
	clrcmd[2] = '1';
	clrcmd[3] = 'K';
	clrcmd[4] = 0;

	poscmd[0] = 0x1b;
	poscmd[1] = '[';
	poscmd[2] = '0';
	poscmd[3] = '0';
	poscmd[4] = ';';
	poscmd[5] = '0';
	poscmd[6] = '0';
	poscmd[7] = 'H';
	poscmd[8] = 0;

	//check to see if this is the first time startup
	uint8_t checksum = 0;
	checksum = eeprom_read_byte((uint8_t *)EETEST);
	checksum += eeprom_read_byte((uint8_t *)EETEST + 1);
	checksum += eeprom_read_byte((uint8_t *)EETEST + 2);
	checksum += eeprom_read_byte((uint8_t *)EETEST + 3);
	if (checksum == 6)//it's not the first time so restore saved settings
	{
		SendString("restoring settings");
		PutChar(13);
		//get some stuff from EEPROM 
		mode = eeprom_read_byte((uint8_t *)EEMODE);
		if (mode > 5)
			mode = 0;

		metric = eeprom_read_byte((uint8_t *)EEUNIT);
		if (metric > 1)
			metric = 0;

		screen = eeprom_read_byte((uint8_t *)EESCREEN);

		mask = eeprom_read_byte((uint8_t *)EEMASK);

		eeprom_read_block (&count, (uint8_t *)EECOUNT, 4);

		eeprom_read_block (&scalefactor, (uint8_t *)EESCALEFACTOR, 4);

		labelf = eeprom_read_byte((uint8_t *)EELABELF);

		labelm = eeprom_read_byte((uint8_t *)EELABELM);

		width = eeprom_read_byte((uint8_t *)EEWIDTH);

		mode0x = eeprom_read_byte((uint8_t *)EEMODE0X);

		mode0y = eeprom_read_byte((uint8_t *)EEMODE0Y);

		mode1x = eeprom_read_byte((uint8_t *)EEMODE1X);

		mode1y = eeprom_read_byte((uint8_t *)EEMODE1Y);

		mode2x = eeprom_read_byte((uint8_t *)EEMODE2X);

		mode2y = eeprom_read_byte((uint8_t *)EEMODE2Y);

		mode3x = eeprom_read_byte((uint8_t *)EEMODE3X);

		mode3y = eeprom_read_byte((uint8_t *)EEMODE3Y);

		mode4x = eeprom_read_byte((uint8_t *)EEMODE4X);

		mode4y = eeprom_read_byte((uint8_t *)EEMODE4Y);


		SetPositionCommand();
	}
	else //first time so set up defaults
	{
		SendString("first run, setting defaults");
		PutChar(13);
		ExecuteCommand('r',0,0,0);
		count = 0;
		eeprom_write_byte((uint8_t *)EETEST,0);
		eeprom_write_byte((uint8_t *)EETEST + 1,1);
		eeprom_write_byte((uint8_t *)EETEST + 2,2);
		eeprom_write_byte((uint8_t *)EETEST + 3,3);
		uint8_t scriptclear[13];
		scriptclear[0] = 0x1b;
		scriptclear[1] = 'Q';
		scriptclear[2] = 0x1b;
		scriptclear[3] = 'R';
		scriptclear[4] = 0x1b;
		scriptclear[5] = '[';
		scriptclear[6] = '8';
		scriptclear[7] = 'v';
		scriptclear[8] = 0x1b;
		scriptclear[9] = '[';
		scriptclear[10] = '1';
		scriptclear[11] = 'v';
		scriptclear[12] = 0;
		//clear the BOB-4 bootscript
		SpiSendString(scriptclear);
	}
	set_count();

	//initial screen refresh
	Refresh();
	while (1)
	{
		//get current count
		get_count();

		//test the switches
		if (!debounce)
		{
			ReadSwitches();
		}
		//refresh the display
		if (mode !=5)
		{
			if (count != lastcount)
			{
				Refresh();
				lastcount = count;
			}
		}
		//check the serial buffer
		if (ReadPointer != WritePointer)
		{
			ParseCommand();
		}
	}
}		
/**********************************************/
